/*
    QJack make you connect with jack soundserver system
    Copyright (C) 2011  Alessandro Siniscalchi <asiniscalchi@gmail.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include "qjackclient.h"
#include <QtCore/QDebug>
#include "jackport_p.h"

namespace QJack {

//! pimpl
class QJackClientPrivate
{
public:
    virtual ~QJackClientPrivate(){};

    bool setFeatures()
    {
        m_nFrames = jack_get_buffer_size(m_client);
        m_sampleRate = jack_get_sample_rate(m_client);

        qDebug() << "QJackClient | client(" << m_clientName << ") nFrames(" << m_nFrames << ") " << "sampleRate(" << m_sampleRate << ")";

        return(true);
    }

    static int process (jack_nframes_t nframes, void *arg)
    {
        //    qDebug() << Q_FUNC_INFO << ": frames = " << nframes;
        const QJackClientPrivate *that = reinterpret_cast<const QJackClientPrivate*>(arg);

        for (int i = 0 ; i < that->m_ports.size() ; i++) {
            JackPort *qJPort = that->m_ports[i];
            // I get the pimpl
            const JackPortPrivate *d = qJPort->d_func();

            void *buffer = jack_port_get_buffer (d->m_port, nframes);
            qJPort->setData(buffer, nframes);
        } // for

        return(0);
    }

    static void jack_shutdown (void *arg)
    {
        qWarning() << Q_FUNC_INFO << ": called ... doing nothing";
    }


    static int jackBufferSizeCallback(jack_nframes_t nframes, void *arg)
    {
        qDebug() << Q_FUNC_INFO << ": nframes = " << nframes;
        return(0);
    }

    bool setCallbacks()
    {
        /* tell the JACK server to call `process()' whenever
             there is work to be done.
          */
        if (jack_set_process_callback (m_client, process, this))
        {
            qWarning() << "jack_set_process_callback : failed";
            return(false);
        }

        if (jack_set_buffer_size_callback(m_client, jackBufferSizeCallback, this)) {
            qWarning() << "jack_set_buffer_size_callback: failed";
            return(false);
        }

        /* tell the JACK server to call `jack_shutdown()' if
           it ever shuts down, either entirely, or if it
           just decides to stop calling us.
        */
        jack_on_shutdown (m_client, jack_shutdown, this);

        return(true);
    }

    bool openConnection()
    {
        m_client = jack_client_open (m_clientName.toAscii().data(), JackNoStartServer , &m_status, "default");
        if (!m_client) {

            if (checkStatus(JackServerFailed)) {
                qWarning() <<"QJackClient | Unable to connect to JACK server";
                return(false);
            }
        }

        if (checkStatus(JackServerStarted)) {
            qDebug() << "QJackClient | can't find the JACK server";
            return(false);
        }
        else
        {
            qWarning() << "QJackClient | JACK server found";
        }

        if (checkStatus(JackNameNotUnique)) {
            m_clientName = jack_get_client_name(m_client);
            qDebug() << Q_FUNC_INFO << "unique name " << m_clientName << " assigned";
        }

        return(true);
    }

    bool checkStatus(JackStatus value)
    {
        return(m_status & value);
    }

    QList<JackPort *> m_ports;
    quint32 m_nFrames;
    quint32 m_sampleRate;
    QString m_clientName;
    jack_client_t *m_client;
    jack_status_t m_status;
}; // class QJackClientPrivate



QJackClient::QJackClient(QString name, QObject *parent): QObject(parent),
    d_ptr(new QJackClientPrivate)
{
    Q_D(QJackClient);

    d->m_clientName = name;
    d->m_client = 0;
}

QJackClient::QJackClient(QJackClientPrivate &dd, QObject *parent):
    QObject(parent), d_ptr(&dd)
{}

QJackClient::~QJackClient()
{
    delete(d_ptr);
}

bool QJackClient::open()
{
    Q_D(QJackClient);

    if (d->m_client)
        return(false);

    // opening a connection with the jack server
    if (!d->openConnection()) return(false);

    // ask for the buffers features
    if (!d->setFeatures()) return(false);

    // setting the callbacks
    if (!d->setCallbacks()) return(false);

    return(true);
}

quint32 QJackClient::nFrames() const
{
    Q_D(const QJackClient);

    return(d->m_nFrames);
}

quint32 QJackClient::sampleRate() const
{
    Q_D(const QJackClient);

    return(d->m_sampleRate);
}

bool QJackClient::activate()
{
    Q_D(QJackClient);

    if (!d->m_client)
    {
        return(false);
    }

    if (d->m_ports.size() == 0)
    {
        qWarning() << "QJackClient | Can't activate a client without ports";
        return(false);
    }

    if (jack_activate (d->m_client)) {
        qWarning() << "QJackClient | cannot activate client";
        return(false);
    }
    return(true);
}

bool QJackClient::deactivate()
{
    Q_D(QJackClient);

    if (!d->m_client) {
        qWarning() << "QJackClient | deactive .. client isn'ty opened";
        return(false);
    }

    if (jack_deactivate(d->m_client)) {
        qWarning() << "QJackClient | deactivation doesn't work";
        return(false);
    }

    return(true);
}

bool QJackClient::close()
{
    Q_D(QJackClient);

    if (!d->m_client) {
        qWarning() << "QJackClient | close .. client isn'ty opened";
        return(false);
    }

    if (jack_client_close(d->m_client)) {
        qWarning() << "QJackClient | closing doesn't work";
        return(false);
    }

    d->m_client = 0;
    return(false);
}

bool QJackClient::addPort(JackPort *port)
{
    if (!port) {
        qWarning() << Q_FUNC_INFO << ": null port";
        return(false);
    }

    Q_D(QJackClient);

    // I get the pimpl
    JackPortPrivate *d_port = port->d_func();

    qDebug() << "QJackClient | trying to add port = " << port->name() << " type = " << d_port->m_type;

    d_port->m_port = jack_port_register (d->m_client, port->name().toAscii().data(),
                                         JACK_DEFAULT_AUDIO_TYPE,
                                         d_port->m_type, 0);

    // check if port was created
    if (!d_port->m_port) {
        qWarning() << "QJackClient | no ports available";
        return(false);

    }

    d->m_ports.append(port);

    return(true);
}

} // namespace

#include "qjackclient.moc"


